/*
  *****************************************************************************

  Copyright 1993 The Regents of the University of California.
  All Rights Reserved.

  Permission to use, copy, modify and distribute this software and its
  documentation for educational, research and non-profit purposes, without
  fee, and without a written agreement is hereby granted, provided that the 
  above copyright notice and the following three paragraphs appear in all 
  copies.

  Permission to incorporate this software into commercial products may be
  obtained from the Office of Technology Licensing, 2150 Shattuck Avenue,
  Suite 510, Berkeley, CA 94720-1620.

  IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
  DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
  LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION,
  EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY
  OF SUCH DAMAGES.

  THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
  AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
  ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATIONS
  TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.

  The author may be contacted via:

  US Mail:  Brian Mirtich
            387 Soda Hall
	    Computer Science Division
	    Department of Electrical Engineering and Computer Science
	    University of California
	    Berkeley, CA 94720

 Phone:     (510) 642-8149

 EMail:     mirtich@cs.berkeley.edu


 Revisions:

 5 Sep 1995  Corrected bugs in face_face() related to setting the
             penetration flag.  Thanks, Ming!

 21 Oct 1996 Corrected bug in computeVertCone.  "vectNormalize(u);"
             becomes "vectNormalize(u, u);".  Thanks to Moritz Breipohl
	     for catching this.

 04 Apr 1997 Corrected the formal parameter lists in Dee() and
             closestToEdge(), and removed many unused local variables
	     throughout the program.  These corrections were all cosmetic; 
	     the behavior should be the same.

 *****************************************************************************
*/



#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include "matvect.h"
#include "cf.h"



/*
  =============================================================================
  
  global variables

  =============================================================================
*/


POLYHEDRON polyhedronLibrary[100];
int polyhedronLibraryCount;


/*
  =============================================================================
  
  miscellaneous routines

  These are general functions which are not specifically related to the 
  distance algorithm, but are called by some of the routines.

  =============================================================================
*/


/* 
  Get next word on the line; skip over spaces and tabs.
   return true if we read a word, false if we hit \n first.
 */

int getWord(fp, s)
     FILE *fp;
     char *s;
{
  char c;

  do fscanf(fp, "%c", &c); while ((c == ' ') || c == '\t');
  if (c == '\n') return 0;
  ungetc(c, fp);
  fscanf(fp, "%s", s);
  return 1;
}


/* return random # from {0, 1, ..., n-1} */

int random2(n)
     int n;
{
  float x;

  x = ((float) rand()) / 0x7FFFFFFF;  /* x = random # in interval [0, 1) */
  return (int) floor(x * n);
}



/*
  =============================================================================
  
  polyhedra and feature housekeeping

  These routines are for accessing the polyhedron library, and performing
  basic operations on polyhedra and their features.

  =============================================================================
*/


/* 
  Add a feature to a feature list.  This is a generic routine callable for
  any type of feature and feature list.
*/

void addFeature(feat, list)
  void *feat;
  FNODE **list;
{
  FNODE *fn;
  
  fn = allocFnode;
  fn->f.any = feat;
  fn->next = *list;
  *list = fn;
}


/* 
  reverse feature list

  Because of the way feature lists are constructed (in a stack-like manner),
  it is sometimes necessary to reverse the order of the list.  This is
  necessary, for example, to make the vertices in a face's vertex list
  appear in the list in CCW order.  This routine reverse any feature list.
*/
  
void reverseFlist(list)
     FNODE **list;
{
  FNODE *last, *fn, *next;

  for (last = NULL, fn = *list; fn; fn = next) {
    next = fn->next;
    fn->next = last;
    last = fn;
  }
  if (last) *list = last;
}

    
/* 
  Get name of an arbitrary feature f, and return in s.  Vertices and faces
  have explicit names; the name of an edge is derived from the names of
  its endpoint vertices (in alphabetical order), separated by a "." .
*/

void featureName(f, s)
     void *f;
     char *s;
{
  int type;

  type = featTag(f);
  if (type == V) strcpy(s, ((VERTEX *)f)->name);
  else if (type == F) strcpy(s, ((FACE *)f)->name);
  else sprintf(s, "%s.%s", ((EDGE *)f)->v1->name, ((EDGE *)f)->v2->name);
}


/*
  Return pointer to a vertex stucture, given the vertex's name and the
  polyhedron where it lives.
*/

VERTEX *findVertex(p, name)
     POLYHEDRON *p;
     char *name;
{
  FNODE *fn;
  
  for (fn = p->verts; fn && strcmp(name, fn->f.v->name); fn = fn->next);
  return (fn) ? fn->f.v : NULL;
}


/*
  Return pointer to an edge stucture, given the names of the edges endpoint
  vertices, and the polyhedron where it lives.  The names of the vertices
  may be passed in either order.
*/

EDGE *findEdge(p, name1, name2)
     POLYHEDRON *p;
     char *name1, *name2;
{
  FNODE *fn;
  
  if (strcmp(name1, name2) > 0) SWAP(name1, name2, char *);
  for (fn = p->edges;
       fn && (strcmp(name1, fn->f.e->v1->name) ||
	      strcmp(name2, fn->f.e->v2->name)); fn = fn->next);
  return (fn) ? fn->f.e : NULL;
}


/*
  Return pointer to a face stucture, given the face's name and the
  polyhedron where it lives.
*/

FACE *findFace(p, name)
     POLYHEDRON *p;
     char *name;
{
  FNODE *fn;
  
  for (fn = p->faces; fn && strcmp(name, fn->f.f->name); fn = fn->next);
  return (fn) ? fn->f.f : NULL;
}


/* 
  return total number of features in a polyhedron
 */

int numFeats(p)
     POLYHEDRON *p;
{
  FNODE *fn;
  int n;

  n = 0;
  for (fn = p->verts; fn; fn = fn->next) n++;
  for (fn = p->edges; fn; fn = fn->next) n++;
  for (fn = p->faces; fn; fn = fn->next) n++;
  return n;
}


/*
  Return the nth feature of a polyhedron.  The features are numbered starting 
  from 0.  Vertices are first, followed by edges, and then faces.
*/

void *nthFeat(p, n)
     POLYHEDRON *p;
     int n;
{
  FNODE *fn;

  for (fn = p->verts; n > 0 && fn; fn = fn->next) n--;
  if (fn) return fn->f.any;
  for (fn = p->edges; n > 0 && fn; fn = fn->next) n--;
  if (fn) return fn->f.any;
  for (fn = p->faces; n > 0 && fn; fn = fn->next) n--;
  return fn->f.any;
}


/* 
  Return a random feature from polyhedron.  There are equal chances that
  a vertex, edge, or face will be returned.  Within each category, all
  features have an equal chance of being selected.
*/
  
void *randFeat(p)
     POLYHEDRON *p;
{
  int type;
  FNODE *fn, *fn2;
  int i, n;

  type = random2(3);
  if (type == 0) fn = p->verts;
  else if (type == 1) fn = p->edges;
  else fn = p->faces;

  /* count number of features in the selected list */
  for (fn2 = fn, n = 0; fn2; n++, fn2 = fn2->next);

  /* pick a random feature */
  n = random2(n);
  for (i = 0; i < n; i++) fn = fn->next;
  return fn->f.any;
}


/* 
  Return radius of a library polyhedron.  The radius of a polyhedron is
  the maximum "norm" of a vertex, where the norm of a vertex is the norm
  of the position vector from the vertex which is rooted at the origin.
*/

REAL polyRad(name)
     char *name;
{
  REAL norm, rad;
  FNODE *fn;
  int i;

  for (i = 0; i < polyhedronLibraryCount && 
              strcmp(polyhedronLibrary[i].name, name) ; i++);
  if (i == polyhedronLibraryCount) {
    printf("***** error:  can't find %s in library\n", name);
    exit(1);
  }
  
  rad = 0.0;
  for (fn = polyhedronLibrary[i].verts; fn; fn = fn->next) {
    norm = vectNorm(fn->f.v->coords);
    if (norm > rad) rad = norm;
  }
  return rad;
}


/*
  Create a new edge between vertices v1 and v2.  Initialize the edge's
  vertex pointers, and it's unit direction vector and length fields.
*/

EDGE *newEdge(v1, v2)
     VERTEX *v1, *v2;
{
  EDGE *e;

  e = allocEdge;
  e->tag = E;
  if (strcmp(v1->name, v2->name) > 0) SWAP(v1, v2, VERTEX *);
  e->v1 = v1;
  e->v2 = v2;
  addFeature(e, &v1->edges);
  addFeature(e, &v2->edges);
  e->fl = e->fr = NULL;
  e->cone = NULL;
  vectSub(v2->coords, v1->coords, e->u);
  e->len = vectNorm(e->u);
  e->u[0] /= e->len;
  e->u[1] /= e->len;
  e->u[2] /= e->len;
  return e;
}


/*
  Read the polyhedron library from the file whose name is passed in.  The 
  library is stored into the global variable polyhedronLibrary.  The global 
  variable polyhedronLibraryCount is also initialized.
*/

int loadPolyhedronLibrary(fname)
     char *fname;
{
  FILE *fp;
  char s[80];
  POLYHEDRON *p;
  VERTEX *v;
  FACE *f;
  EDGE *e, *e1, *e2;
  VERTEX *start, *last;
  FNODE *fn;
  int cont;
  int nameCounter;
  int n;

  printf("\nloading polyhedron library:\n");
  n = 0;
  fp = fopen(fname, "r");

  while(1) {

    do fscanf(fp, "%s", s); while (!feof(fp) && strcmp(s, "polyhedron"));
    if (feof(fp)) break;

    p = &polyhedronLibrary[n++];
    fscanf(fp, "%s", p->name);
    printf("%s\n", p->name);
    mat4copy(mat4I, p->pose);
    p->verts = p->edges = p->faces = NULL;

    /* read vertices */
    while (1) {
      fscanf(fp, "%s", s);
      if (s[0] == '*') break;
      v = allocVertex;
      v->tag = V;
      strcpy(v->name, s);
      v->cone = NULL;
      if (sizeof v->coords[0] == sizeof(double)) 
        fscanf(fp, "%lf %lf %lf", &v->coords[0], &v->coords[1], &v->coords[2]);
      else
        fscanf(fp, "%f %f %f", &v->coords[0], &v->coords[1], &v->coords[2]);
      v->edges = NULL;
      addFeature(v, &p->verts);
    }

    /* read faces */
    nameCounter = 0;
    while (1) {
      fscanf(fp, "%s", s);
      if (s[0] == '*') break;
      if (s[0] == '-') sprintf(s, "#F_%d", ++nameCounter);
      f = allocFace;
      f->tag = F;
      strcpy(f->name, s);
      f->verts = f->edges = NULL;
      f->cone = NULL;
      addFeature(f, &p->faces);
      getWord(fp, s);
      start = last = findVertex(p, s);
      do {
	if (cont = getWord(fp, s)) v = findVertex(p, s);
	else v = start;
	addFeature(v, &f->verts);
	if (e = findEdge(p, last->name, v->name)) {
	  if (e->fl) e->fr = f;
	  else e->fl = f;
	}
	else {
	  e = newEdge(last, v);
	  if (strcmp(last->name, v->name) < 0) e->fl = f;
	  else e->fr = f;
	  addFeature(e, &p->edges);
	}
	addFeature(e, &f->edges);
	last = v;
      } while (cont);
      /* compute face plane coeffs (& outward normal)  */
      e1 = f->edges->f.e;
      e2 = f->edges->next->f.e;
      /* the "not" below is necessary cuz we haven't reversed the
	 vertex list to its proper CCW order yet; right now it's in CW order */
      if (!(e1->v1 == e2->v2 || e1->v2 == e2->v1))
	vectXprod(e1->u, e2->u, f->plane);
      else vectXprod(e2->u, e1->u, f->plane);
      vectNormalize(f->plane, f->plane);
      f->plane[3] = - vectDotProd(f->plane, f->verts->f.v->coords);
    }

    /* clean up lists */
    for (fn = p->verts; fn; fn = fn->next) reverseFlist(&fn->f.v->edges);
    for (fn = p->faces; fn; fn = fn->next) {
      reverseFlist(&fn->f.f->verts->next);
      reverseFlist(&fn->f.f->edges);
    }
    reverseFlist(&p->verts);
    reverseFlist(&p->edges);
    reverseFlist(&p->faces);

    /* build the voronoi regions for the polyhedron */
    buildCones(p);

  }
  
  fclose(fp);
  polyhedronLibraryCount = n;
  printf("%d polyhedra in library\n\n", n);
  return n;

}


/*
  create polyhedron

  Once the polyhedron library has been initialized, we can create instances
  of the various polyhedron in the library.  Any number of instances of the 
  same polyhedron may exist, at different positions (poses) in space.  We 
  create a polyhedron instance by specifying the name of the reference
  polyhedron in the library (e.g. "cube"), and the name for this particular
  instance (e.g. "cube-1").  The former name is required; the latter one
  is optional, and can be NULL.  A pointer to the newly instantiated polyhedron
  is returned.  The pose matrix is initialized to the identity.  Note that
  all instances of the same library polyhedron share the same vertex, edge,
  and face lists and voronoi structure.
*/

POLYHEDRON *createPolyhedron(libName, name)
     char *libName, *name;
{
  POLYHEDRON *newP;
  int i;

  for (i = 0; i < polyhedronLibraryCount && 
              strcmp(polyhedronLibrary[i].name, libName) ; i++);
  if (i == polyhedronLibraryCount) {
    printf("***** error:  can't find %s in library\n", libName);
    return NULL;
  }

  newP = allocPolyhedron;
  strcpy(newP->name, name);
  mat4copy(mat4I, newP->pose);
  newP->verts = polyhedronLibrary[i].verts;
  newP->edges = polyhedronLibrary[i].edges;
  newP->faces = polyhedronLibrary[i].faces;

  return newP;
}


/*
  print out all the features of a polyhedron
*/

void dumpPolyhedron(p)
     POLYHEDRON *p;
{
  FNODE *fn, *fn2;
  char s[80];
  
  printf("polyhedron %s ===================\n", p->name);

  printf("vertices:\n");
  for (fn = p->verts; fn; fn = fn->next) {
    printf("%-10s  (%+6.2f, %+6.2f, %+6.2f)\n", fn->f.v->name, 
	   fn->f.v->coords[0], fn->f.v->coords[1], fn->f.v->coords[2]);
    printf("  edges:");
    for (fn2 = fn->f.v->edges; fn2; fn2 = fn2->next)
      printf("  %s.%s", fn2->f.e->v1->name, fn2->f.e->v2->name);
    printf("\n");
  }

  printf("edges:\n");
  for (fn = p->edges; fn; fn = fn->next) {
    sprintf(s, "%s.%s", fn->f.e->v1->name, fn->f.e->v2->name);
    printf("%-15s   l=%-10s r=%-10s\n", 
	   s, fn->f.e->fl->name, fn->f.e->fr->name);
    printf("  l = %+7.2f  u = (%+6.3f, %+6.3f, %+6.3f)\n", fn->f.e->len, 
	   fn->f.e->u[0], fn->f.e->u[1], fn->f.e->u[2]);
  }

  printf("faces:\n");
  for (fn = p->faces; fn; fn = fn->next) {
    printf("%-10s  (%+6.2f, %+6.2f, %+6.2f)  %+6.2f\n", fn->f.f->name, 
	   fn->f.f->plane[0], fn->f.f->plane[1],
	   fn->f.f->plane[2], fn->f.f->plane[3]);
    printf("  vertices:");
    for (fn2 = fn->f.f->verts; fn2; fn2 = fn2->next)
      printf("  %s", fn2->f.v->name);
    printf("\n  edges:");
    for (fn2 = fn->f.f->edges; fn2; fn2 = fn2->next)
      printf("  %s.%s", fn2->f.e->v1->name, fn2->f.e->v2->name);
    printf("\n");
  }
}



/*
  =============================================================================
  
  voronoi region routines

  These routines are used for building the voronoi region structures
  of polyhedra, and for testing for membership of points in these regions.

  The three membership routines are vertConeChk, edgeConeChk, and
  faceConeChk, for testing membership of points within vertex, edge, and
  face voronoi regions respectively.  Actually, the faceConeChk only checks
  against the side planes of the face's voronoi region; the base plane 
  must be checked explicitly.  "Points" are simply arrays of 3 REALs.

  All three membership routines are passed the feature whose cone is being
  checked as well as the point.  These routines return true if the point was
  in the cone, and false otherwise.  In the latter case, the feature pointer
  passed in will be updated with a new feature (of a different type) which
  neighbored the old feature, corresponding to the plane which was violated.
  We do not update the feature based on the first plane which is found to
  be violated, but rather on the plane which is "most violated."  The
  amount of violation is measured by the distance of the point from the
  plane, assuming the point is on the negative side of the plane.  A point
  on the positive side does not violated the plane.  Hence, even if we
  find a plane which is violated, we continue to cycle through all planes
  of the cone, looking for the one which is most violated.

  The voronoi regions are "sticky" in that we make comparisons against
  -EPSILON rather than zero.  This provides hysteresis to avoid cycling.
  The voronoi planes are oriented so that the "signed distance" of a point 
  to a voronoi plane is positive if the point is on the region side of the 
  plane, and negative if the point lies outside the region.

  N.B.  The order in which the edge cone planes are built is important!
  We add the planes corresponding to vertices (the edge's endpoints) first,
  followed by the planes corresponding to the neighboring faces.  This
  puts the face planes in the front of the edge's cone list.  We depend
  on this order in the edge_face routine when we're checking that the the
  faces normal lies between the edge's neigboring faces' normals.

  =============================================================================
*/


/* 
  Add a plane node to a cone list.  A cone list is a list of planes describing
  the voronoi region of a vertex or an edge, or the side planes of the
  voronoi region of a face.
*/

void addPlane(pn, cone)
     PNODE *pn;
     PNODE **cone;
{
  pn->next = *cone;
  *cone = pn;
}


/*
  flip plane

  Planes have an orientation, since we perform checks to see if points lie
  on the proper side of the plane.  flipPlane negates the four parameters
  describing the source plane, to produce a destination plane which is
  identical to the source, except with opposite orientation.
*/

void flipPlane(src, dest)
     VECT4 src, dest;
{
  dest[0] = - src[0];
  dest[1] = - src[1];
  dest[2] = - src[2];
  dest[3] = - src[3];
}


/*
  tweak plane normal

  Rotate a plane normal epsilon radians about a given axis u.  This 
  should be done before the constant term of the plane vector is
  computed, since this term is based on the plane normal.
  u must be a unit vector!
*/

void tweakPlaneNorm(u, eps, nOrig, nTweak)
     VECT3 u, nOrig, nTweak;
     REAL eps;
{
  MAT3 R;
  REAL c, s, v;

  /* compute rotation matrix for rotating 3-vectors an angle eps about u */
  sincos(eps, &s, &c);
  v = 1.0 - c;
  R[0][0] = u[0] * u[0] * v + c;
  R[1][1] = u[1] * u[1] * v + c;
  R[2][2] = u[2] * u[2] * v + c;
  R[0][1] = u[0] * u[1] * v - u[2] * s;
  R[0][2] = u[0] * u[2] * v + u[1] * s;
  R[1][0] = u[1] * u[0] * v + u[2] * s;
  R[1][2] = u[1] * u[2] * v - u[0] * s;
  R[2][0] = u[2] * u[0] * v - u[1] * s;
  R[2][1] = u[2] * u[1] * v + u[0] * s;

  xform3(R, nOrig, nTweak);
}
  

/*
  Compute the voronoi region of a vertex.  This creates a list of plane
  nodes (one for each incident edge) and points the vertex's cone field
  to this list.
*/

void computeVertCone(v)
     VERTEX *v;
{
  PNODE *pn;
  FNODE *fn;
  EDGE *e;
  VECT3 u, tmpV;

  for (fn = v->edges; fn; fn = fn->next) {
    e = fn->f.e;
    /* compute vector about which to rotate vertex planes REWRITE
     this vector points toward the right of edge */
    vectAdd(e->fl->plane, e->fr->plane, tmpV);
    vectXprod(e->u, tmpV, u);
    vectNormalize(u, u);  /* bug fix:  Moritz Breipohl, 21.10.96 */

    /* construct vertex plane */
    pn = allocPnode;
    if (e->v1 == v) {
      tweakPlaneNorm(u, VERT_FLARE, e->u, pn->plane);
      vectNeg(pn->plane, pn->plane);
    }
    else tweakPlaneNorm(u, -VERT_FLARE, e->u, pn->plane);
    pn->plane[3] = - vectDotProd(v->coords, pn->plane);
    pn->nbr = e;
    addPlane(pn, &v->cone);

    /* construct edge plane */
    pn = allocPnode;
    if (e->v1 == v) tweakPlaneNorm(u, EDGE_V_FLARE, e->u, pn->plane);
    else {
      tweakPlaneNorm(u, -EDGE_V_FLARE, e->u, pn->plane);
      vectNeg(pn->plane, pn->plane);
    }
    pn->plane[3] = - vectDotProd(v->coords, pn->plane);
    pn->nbr = v;
    addPlane(pn, &e->cone);
  }
}


/*
  Compute the voronoi region of a face.  This creates a list of plane
  nodes (one for each edge of the face) and points the faces's cone field
  to this list.  The final plane of the voronoi region is the plane of
  the face itself.  This plane is stored explicitly when the face is
  created.
*/

void computeFaceCone(f)
     FACE *f;
{
  PNODE *pn;
  EDGE *e;
  VERTEX *v;
  FNODE *fnE, *fnV;
  VECT3 norm;
  
  /* we only do side planes; 
     base plane is already stored explicitly in f->plane */
  for (fnE = f->edges, fnV = f->verts; fnE; fnE = fnE->next, fnV = fnV->next) {
    e = fnE->f.e;
    v = fnV->f.v;
    /*
      (e->v1 = v) <==> e goes CCW around f; f is left face of e
      (e->v2 = v) <==> e goes CW around f; f is right face of e
    */

    /* construct face plane */
    vectXprod(f->plane, e->u, norm);
    pn = allocPnode;
    if (e->v1 == v) tweakPlaneNorm(e->u, FACE_FLARE, norm, pn->plane);
    else {
      tweakPlaneNorm(e->u, -FACE_FLARE, norm, pn->plane);
      vectNeg(pn->plane, pn->plane);
    }
    pn->plane[3] = - vectDotProd(v->coords, pn->plane);
    pn->nbr = e;
    addPlane(pn, &f->cone);

    /* construct edge plane */
    pn = allocPnode;
    if (e->v1 == v) {
      tweakPlaneNorm(e->u, EDGE_F_FLARE, norm, pn->plane);
      vectNeg(pn->plane, pn->plane);
    }
    else tweakPlaneNorm(e->u, -EDGE_F_FLARE, norm, pn->plane);
    pn->plane[3] = - vectDotProd(v->coords, pn->plane);
    pn->nbr = f;
    addPlane(pn, &e->cone);
  }
}


/*
  build the voronoi region structure for an entire polyhedron
*/

void buildCones(p)
     POLYHEDRON *p;
{
  FNODE *fn;

  for (fn = p->verts; fn; fn = fn->next) computeVertCone(fn->f.v);
  for (fn = p->faces; fn; fn = fn->next) computeFaceCone(fn->f.f);
}


/*
  print out the voronoi region structure for an entire polyhedron
*/

void dumpCones(p)
     POLYHEDRON *p;
{
  FNODE *fn;
  PNODE *pn;

  printf("cone structure of polyhedron %s ===============\n", p->name);

  for (fn = p->verts; fn; fn = fn->next) {
    printf("*** vertex %s\n", fn->f.v->name);
    for (pn = fn->f.v->cone; pn; pn = pn->next)
      printf("%+7.3f %+7.3f %+7.3f %+7.3f => %s.%s\n",
	     pn->plane[0], pn->plane[1], pn->plane[2], pn->plane[3],
	     ((EDGE *)pn->nbr)->v1->name, 
	     ((EDGE *)pn->nbr)->v2->name);
  }

  for (fn = p->edges; fn; fn = fn->next) {
    printf("*** edge %s.%s\n", fn->f.e->v1->name, fn->f.e->v2->name);
    for (pn = fn->f.e->cone; pn; pn = pn->next)
      printf("%+7.3f %+7.3f %+7.3f %+7.3f => %s\n",
	     pn->plane[0], pn->plane[1], pn->plane[2], pn->plane[3],
	     (featTag(pn->nbr) == V) ? 
	     ((VERTEX *)pn->nbr)->name : ((FACE *)pn->nbr)->name);
  }

  for (fn = p->faces; fn; fn = fn->next) {
    printf("*** face %s\n", fn->f.f->name);
    for (pn = fn->f.f->cone; pn; pn = pn->next) {
      printf("%+7.3f %+7.3f %+7.3f %+7.3f => ", 
	     pn->plane[0], pn->plane[1], pn->plane[2], pn->plane[3]);
      if (pn->nbr) printf("%s.%s\n",
			  ((EDGE *)pn->nbr)->v1->name, 
			  ((EDGE *)pn->nbr)->v2->name);
      else printf("----\n");
    }
  }

}


/* 
  Return true if a point lies within vertex v's voronoi region, else false.
  if update is true, update v to a neighboring edge in the latter case.
*/

int vertConeChk(v, point, update)
     VERTEX **v;
     VECT3 point;
     int update;
{
  PNODE *pn;
  REAL min, dot;
  void *feat;

  min = INFINITY;
  for (pn = (*v)->cone; pn; pn = pn->next)
    if ((dot = planeDist(pn->plane, point)) < min) {
      min = dot;
      feat = pn->nbr;
    }
  if (min > -CONE_EPS) return 1;
  if (update) *v = feat;
  return 0;
}


/* 
  Return true if a point lies within edge e's voronoi region, else false.
  if update is true, update e to a neighboring vertex or face in the latter 
  case.
*/

int edgeConeChk(e, point, update)
     EDGE **e;
     VECT3 point;
     int update;
{
  PNODE *pn;
  REAL min, dot;
  void *feat;

  min = INFINITY;
  for (pn = (*e)->cone; pn; pn = pn->next)
    if ((dot = planeDist(pn->plane, point)) < min) {
      min = dot;
      feat = pn->nbr;
    }
  if (min >= -CONE_EPS) return 1;
  if (update) *e = feat;
  return 0;
}


/* 
  Return true if a point lies within face f's voronoi region, else false.
  If update is true, update f to a neighboring edge in the latter case.
*/

int faceConeChk(f, point, update)
     FACE **f;
     VECT3 point;
     int update;
{
  PNODE *pn;
  REAL min, dot;
  void *feat;

  min = INFINITY;
  for (pn = (*f)->cone; pn; pn = pn->next)
    if ((dot = planeDist(pn->plane, point)) < min) {
      min = dot;
      feat = pn->nbr;
    }
  if (min >= -CONE_EPS) return 1;
  if (update) *f = feat;
  return 0;
}



/*
  =============================================================================

  quick and dirty distance functions and closestToX functions

  Occasionally it is necessary to find the closest feature on polyhedron A
  to a particular feature on polyhedron B.  We sometimes have to enumerate
  over all features on A.  This happens, for instance, when a vertex
  on B lies in the "negative voronoi region" of a face on A.

  To perform such operations, we use quick and dirty distance functions.
  These are seven functions which return the distance between two pairs
  of features of a given type.  They are quick and dirty because they
  sometimes return infinity rather than an actual distance.  For instance,
  the distance between two faces is not even calculated unless the faces
  are parallel, for if they are not, we know some some other feature
  pair (possibly involving one of the faces) is closer.  Thus Dff returns
  infinity if the faces aren't parallel.

  These functions are named "Dxy" where x and y are either "v", "e", or
  "f", for vertex, edge or face.  Thus, Dev returns the distance between
  an edge and a vertex. We use these routines when searching over polyhedron
  A for the closest feature to a particular feature on B.  For this reason,
  it is most efficient to transform the feature on B to polyhedron A's
  frame, rather than transforming every feature on A to B's frame.  In
  most of these routines, it is thus assumed that the second feature passed 
  in (from polyhedron B) has already been transformed to the frame of the
  first feature's polyhedron (polyhedron A).  An exception is Dff (see below).

  The quick and dirty distance functions are called by the three closestToX
  functions:  closestToVert, closestToEdge, and closestToFace.  ClosestToVert
  finds the closest feature on A to a vertex on B by first transforming
  the vertex on B into A's frame, and making repeated calls to Dvv, Dev, and
  Dfv.  Likewise, closestToEdge finds the closest feature on A to an edge
  on B by transforming the edge to A's frame and calling Dve, Dee, Dfe while
  enumerating over all features of A.  closestToFace works a little 
  differently, since it is more difficult to transform an entire face into
  another frame.  It still enumerates over all features on A, looking for
  the closest feature to a face on B, however if the current feature being
  examined on A is a vertex or edge, this feature is transformed into B's
  frame (rather than transforming the face on B into A's frame).  Hence,
  we still make calls to Dfv and Dfe (Dvf and Def don't exist).  If the
  current feature on A is a face, we call Dff, which assumes each face is
  in its own frame.  This slight inefficiency in the closestToFace case is
  no big deal, due to the rarity of calling this function.

  Also note in the closestToX functions, we first enumerate over faces, then
  edges, then vertices.  This gives priority to features of higher dimension,
  which is what we want.  For example, if a face and an edge on A are both 
  closest to the fixed feature on B, we'll return the face.

  =============================================================================
*/


/*
  Return distance between two vertices; the latter has been xformed
  into the frame of the former.
*/

REAL Dvv(v1, v2)
     VERTEX *v1, *v2;
{
  VECT3 offset;

  vectSub(v1->coords, v2->xcoords, offset);

  return vectNorm(offset);
}


/*
  Return distance between an edge and a vertex xformed into the edge's frame.
*/

REAL Dev(e, v)
     EDGE *e;
     VERTEX *v;
{
  VECT3 w, offset;
  REAL lambda;
  
  vectSub(v->xcoords, e->v1->coords, w);
  lambda = vectDotProd(w, e->u);
  if (lambda < 0.0) lambda = 0.0;
  else if (lambda > e->len) lambda = e->len;
  displacePoint(e->v1->coords, e->u, lambda, w);
  vectSub(v->xcoords, w, offset);

  return vectNorm(offset);
}


/*
  Return distance between a face and a vertex xformed into the face's frame,
  or possibly infinity.
*/

REAL Dfv(f, v)
     FACE *f;
     VERTEX *v;
{
  REAL dist;
  PNODE *pn;

  for (pn = f->cone; pn; pn = pn->next)
    if (planeDist(pn->plane, v->xcoords) < 0.0) return INFINITY;
  dist = planeDist(f->plane, v->xcoords);
  if (dist < 0.0) return INFINITY;

  return dist;
}


/*
  Return distance between a vertex and an edge xformed into the vertex's frame.
*/
  
REAL Dve(v, e)
     VERTEX *v;
     EDGE *e;
{
  VECT3 w, offset;
  REAL lambda;
  
  vectSub(v->coords, e->v1->xcoords, w);
  lambda = vectDotProd(w, e->xu);
  if (lambda < 0.0) lambda = 0.0;
  else if (lambda > e->len) lambda = e->len;
  displacePoint(e->v1->xcoords, e->xu, lambda, w);
  vectSub(v->coords, w, offset);

  return vectNorm(offset);
}


/*
  Return distance between two edges; the latter has been xformed into
  the frame of the former.
*/

REAL Dee(e1, e2)
     EDGE *e1, *e2;
{
  VECT3 cp1, cp2, offset;

  edgeCPs(&e2, &e1, cp2, cp1);
  vectSub(cp2, cp1, offset);

  return vectNorm(offset);
}


/*
  Return distance between a face and an edge xformed into the face's frame,
  or possibly infinity.
*/

REAL Dfe(f, e)
     FACE *f;
     EDGE *e;
{
  REAL h1, h2;
  REAL min, max;
  VECT3 minCut, maxCut;

  if (!polygonCut(f, e, &min, &max)) return INFINITY;
  displacePoint(e->v1->xcoords, e->xu, min, minCut);
  displacePoint(e->v1->xcoords, e->xu, max, maxCut);

  h1 = planeDist(f->plane, minCut);
  h2 = planeDist(f->plane, maxCut);

  if (h1 < 0.0 || h2 < 0.0) return INFINITY;
  return (h1 < h2) ? h1 : h2;

}


/*
  Return distance between two faces, or possibly infinity.  The faces are
  each assumed to be in their own frames, and T12 and T21 are the
  xformation matrices from f1's frame to f2's frame and vice-versa, resp.

  If the faces are parallel, then we return the distance of a vertex on
  one face to the plane of the other.  However, we do this computation both
  ways and take the minimum distance.  For the vertex of a huge face might
  appear to be quite far off the plane of a tiny one due to limited precision.
*/

REAL Dff(f1, f2, T12, T21)
     FACE *f1, *f2;
     MAT4 T12, T21;
{
  REAL k;
  REAL minDist, dist;
  VECT3 w;
  FNODE *fn;
  EDGE *e;

  xformVect(T21, f2->plane, w);
  k = vectDotProd(f1->plane, w);
  if (fabs(k) < 1.0 - EPSILON) return INFINITY;

  /* At this point we know faces are || */

  minDist = INFINITY;

  /* Find closest edge on f1 to f2 */
  for (fn = f1->edges; fn; fn = fn->next) {
    e = fn->f.e;
    xformEdge(T12, e);
    dist = Dfe(f2, e);
    if (dist < minDist) minDist = dist;
  }

  /* now test edges of f2 against f1 */
  for (fn = f2->edges; fn; fn = fn->next) {
    e = fn->f.e;
    xformEdge(T21, e);
    dist = Dfe(f1, e);
    if (dist < minDist)  minDist = dist;
  }

  return minDist;
}


/*
  Return closest feature on polyhedron p to the vertex v.  Tvp is the
  xformation matrix from v's frame to p's frame.
*/

void *closestToVert(v, p, Tvp)
     VERTEX *v;
     POLYHEDRON *p;
     MAT4 Tvp;
{
  REAL dist, minDist;
  void *feat;
  FNODE *fn;

  /* xform vertex to frame of polyhedron p */
  xformVert(Tvp, v);  
  minDist = INFINITY;

  for (fn = p->faces; fn; fn = fn->next) {
    dist = Dfv(fn->f.f, v);
    if (dist < minDist) {
      minDist = dist;
      feat = fn->f.f;
    }
  }
  for (fn = p->edges; fn; fn = fn->next) {
    dist = Dev(fn->f.e, v);
    if (dist < minDist) {
      minDist = dist;
      feat = fn->f.e;
    }
  }
  for (fn = p->verts; fn; fn = fn->next) {
    dist = Dvv(fn->f.v, v);
    if (dist < minDist) {
      minDist = dist;
      feat = fn->f.v;
    }
  }
  return feat;
}

  
/*
  Return closest feature on polyhedron p to the edge e.  Tep is the
  xformation matrix from e's frame to p's frame.
*/

void *closestToEdge(e, p, Tep)
     EDGE *e;
     POLYHEDRON *p;
     MAT4 Tep;
{
  REAL dist, minDist;
  void *feat;
  FNODE *fn;

  /* xform edge to frame of polyhedron p */
  xformEdge(Tep, e);
  minDist = INFINITY;

  for (fn = p->faces; fn; fn = fn->next) {
    dist = Dfe(fn->f.f, e);
    if (dist < minDist) {
      minDist = dist;
      feat = fn->f.f;
    }
  }
  for (fn = p->edges; fn; fn = fn->next) {
    dist = Dee(fn->f.e, e);
    if (dist < minDist) {
      minDist = dist;
      feat = fn->f.e;
    }
  }
  for (fn = p->verts; fn; fn = fn->next) {
    dist = Dve(fn->f.v, e);
    if (dist < minDist) {
      minDist = dist;
      feat = fn->f.v;
    }
  }
  return feat;
}
  

/*
  Return closest feature on polyhedron p to the face f.  Tfp is the
  xformation matrix from f's frame to p's frame, and Tpf is the
  inverse transformation.
*/

void *closestToFace(f, p, Tfp, Tpf)
     FACE *f;
     POLYHEDRON *p;
     MAT4 Tfp, Tpf;
{
  REAL dist, minDist;
  void *feat;
  FNODE *fn;

  minDist = INFINITY;

  for (fn = p->faces; fn; fn = fn->next) {
    dist = Dff(fn->f.f, f, Tpf, Tfp);
    if (dist < minDist) {
      minDist = dist;
      feat = fn->f.f;
    }
  }
  for (fn = p->edges; fn; fn = fn->next) {
    xformEdge(Tpf, fn->f.e);
    dist = Dfe(f, fn->f.e);
    if (dist < minDist) {
      minDist = dist;
      feat = fn->f.e;
    }
  }
  for (fn = p->verts; fn; fn = fn->next) {
    xformVert(Tpf, fn->f.v);
    dist = Dfv(f, fn->f.v);
    if (dist < minDist) {
      minDist = dist;
      feat = fn->f.v;
    }
  }
  return feat;
}
  


/*
  =============================================================================

  miscellaneous distance support functions

  These functions are called by the six basic feature pair checks (i.e.
  vertex_vertex, vertex_edge, etc.)  They perform functions such as 
  computing the closest edge on a face to a given edge, or determining
  if two faces overlap.

  =============================================================================
*/


/* 
  compute edge closest points

  This routine computes closest points between edges e1 and e2, returning
  them in cp1 and cp2.  All possible degenerecies are handled (I think).
  In cases where the closest points are not unique (this can sometimes
  happen with parallel edges), individual closest points are still 
  selected.  In these cases, we try to choose sensibly, for instance
  choosing the midpoint of the portion of the edge which is closest to
  the other edge.  We assume e1 has been transformed into e2's frame,
  and both cp1 and cp2 are computed relative to e2's frame.
*/

void edgeCPs(e1, e2, cp1, cp2)
     EDGE **e1, **e2;
     VECT3 cp1, cp2;
{
  REAL k, lambda, lambda1, lambda2;
  VECT3 offset, w, y;
  REAL dot12, dot21;
  REAL h, t;
  REAL dist, minDist;
  
  k = vectDotProd((*e1)->xu, (*e2)->u);
  vectSub((*e2)->v1->coords, (*e1)->v1->xcoords, offset);

  if (fabs(k) > 1.0 - EPSILON) { 
    /* case I:  lines || or anti-|| */
    k = (k > 0.0) ? 1.0 : -1.0;
    lambda1 = lambda2 = INFINITY;
    vectSub((*e1)->v1->xcoords, (*e2)->v1->coords, w);
    dot12 = vectDotProd(w, (*e2)->u);
    dot21 = - vectDotProd(w, (*e1)->xu);

    t = dot12;
    h = dot12 + k * (*e1)->len;
    if (t <= 0.0) {
      if (h <= 0.0) lambda2 = 0.0;
      else if (h >= (*e2)->len) lambda2 = (*e2)->len / 2;
    }
    else if (t >= (*e2)->len) {
      if (h <= 0.0) lambda2 = (*e2)->len / 2;
      else if (h >= (*e2)->len) lambda2 = (*e2)->len;
    }
    
    t = dot21;
    h = dot21 + k * (*e2)->len;
    if (t <= 0.0) {
      if (h <= 0.0) lambda1 = 0.0;
      else if (h >= (*e1)->len) lambda1 = (*e1)->len / 2;
    }
    else if (t >= (*e1)->len) {
      if (h <= 0.0) lambda1 = (*e1)->len / 2;
      else if (h >= (*e1)->len) lambda1 = (*e1)->len;
    }

    if (lambda1 != (REAL) INFINITY && lambda2 == (REAL) INFINITY) 
      lambda2 = dot12 + k * lambda1;
    else if (lambda2 != (REAL) INFINITY && lambda1 == (REAL) INFINITY)
      lambda1 = dot21 + k * lambda2;
    else if (lambda1 == (REAL) INFINITY) {  /* n.b. lambda2 also == INFINITY */
      if (t < 0.0) lambda1 = h / 2;
      else if (t > (*e1)->len) lambda1 = (h + (*e1)->len) / 2;
      else if (h < 0.0) lambda1 = t / 2;
      else lambda1 = (t + (*e1)->len) / 2;
      lambda2 = dot12 + k * lambda1;
    }

    /* compute cp's based on lambdas */
  
    if (lambda1 <= 0.0) vectCopy((*e1)->v1->xcoords, cp1);
    else if (lambda1 >= (*e1)->len) vectCopy((*e1)->v2->xcoords, cp1);
    else displacePoint((*e1)->v1->xcoords, (*e1)->xu, lambda1, cp1);
    
    if (lambda2 <= 0.0) vectCopy((*e2)->v1->coords, cp2);
    else if (lambda2 >= (*e2)->len) vectCopy((*e2)->v2->coords, cp2);
    else displacePoint((*e2)->v1->coords, (*e2)->u, lambda2, cp2);

  }
  
  else {
    /* case II:  lines not || */
    vectScale((*e2)->u, w, k);
    vectSub((*e1)->xu, w, w);
    lambda1 = vectDotProd(w, offset) / (1 - k * k);
    vectScale((*e1)->xu, w, k);
    vectSub((*e2)->u, w, w);
    lambda2 = - vectDotProd(w, offset) / (1 - k * k);

    if (lambda1 >= 0.0 && lambda1 <= (*e1)->len &&
	lambda2 >= 0.0 && lambda2 <= (*e2)->len) {
      displacePoint((*e1)->v1->xcoords, (*e1)->xu, lambda1, cp1);
      displacePoint((*e2)->v1->coords, (*e2)->u, lambda2, cp2);
      return;
    }

    /* (lambda1, lambda2) not in allowable region; check boundaries */

    minDist = INFINITY;
    if (lambda1 < 0.0) {
      /* check boundary:  lambda1 = 0 */
      vectSub((*e1)->v1->xcoords, (*e2)->v1->coords, w);
      lambda = vectDotProd(w, (*e2)->u);
      if (lambda < 0.0) lambda = 0.0;
      else if (lambda > (*e2)->len) lambda = (*e2)->len;
      displacePoint((*e2)->v1->coords, (*e2)->u, lambda, w);
      vectSub((*e1)->v1->xcoords, w, y);
      dist = vectNorm(y);
      if (dist < minDist) {
	minDist = dist;
	vectCopy((*e1)->v1->xcoords, cp1);
	vectCopy(w, cp2);
      }
    }
    else if (lambda1 > (*e1)->len) {
      /* check boundary:  lambda1 = length1 */
      vectSub((*e1)->v2->xcoords, (*e2)->v1->coords, w);
      lambda = vectDotProd(w, (*e2)->u);
      if (lambda < 0.0) lambda = 0.0;
      else if (lambda > (*e2)->len) lambda = (*e2)->len;
      displacePoint((*e2)->v1->coords, (*e2)->u, lambda, w);
      vectSub((*e1)->v2->xcoords, w, y);
      dist = vectNorm(y);
      if (dist < minDist) {
	minDist = dist;
	vectCopy((*e1)->v2->xcoords, cp1);
	vectCopy(w, cp2);
      }
    }
    
    if (lambda2 < 0.0) {
      /* check boundary:  lambda2 = 0 */
      vectSub((*e2)->v1->coords, (*e1)->v1->xcoords, w);
      lambda = vectDotProd(w, (*e1)->xu);
      if (lambda < 0.0) lambda = 0.0;
      else if (lambda > (*e1)->len) lambda = (*e1)->len;
      displacePoint((*e1)->v1->xcoords, (*e1)->xu, lambda, w);
      vectSub((*e2)->v1->coords, w, y);
      dist = vectNorm(y);
      if (dist < minDist) {
	minDist = dist;
	vectCopy((*e2)->v1->coords, cp2);
	vectCopy(w, cp1);
      }
    }
    else if (lambda2 > (*e2)->len) {
      /* check boundary:  lambda2 = length2 */
      vectSub((*e2)->v2->coords, (*e1)->v1->xcoords, w);
      lambda = vectDotProd(w, (*e1)->xu);
      if (lambda < 0.0) lambda = 0.0;
      else if (lambda > (*e1)->len) lambda = (*e1)->len;
      displacePoint((*e1)->v1->xcoords, (*e1)->xu, lambda, w);
      vectSub((*e2)->v2->coords, w, y);
      dist = vectNorm(y);
      if (dist < minDist) {
	minDist = dist;
	vectCopy((*e2)->v2->coords, cp2);
	vectCopy(w, cp1);
      }
    }
  }
}


/*
  closest edge or vertex on face

  Return the edge or vertex of face f which is closest to edge e.
  We assume e has been transformed to f's frame.
*/

void *closestEdgeOrVertOnFace(f, e)
     FACE *f;
     EDGE *e;
{
  FNODE *fn;
  REAL dist, minDist;
  EDGE *e2, *closest;
  VECT3 cp1, cp2, tmp;

  minDist = INFINITY;
  for (fn = f->edges; fn; fn = fn->next) {
    e2 = fn->f.e;
    edgeCPs(&e, &e2, cp1, cp2);
    vectSub(cp2, cp1, tmp);
    dist = vectNorm(tmp);
    if (dist < minDist) {
      minDist = dist;
      if (vectEqual(cp2, e2->v1->coords)) closest = (EDGE *) e2->v1;
      else if (vectEqual(cp2, e2->v2->coords)) closest = (EDGE *) e2->v2;
      else closest = e2;
    }
  }
  return closest;
}


/*
  Find the closest pair of edges between faces f1 and f2; return them
  in closest1 & closest2.  T12 is the transformation matrix from f1's
  frame to f2's frame.  This is just a dumb n^2 algorithm which loops
  through all pairs of edges.

  A minor complication is that we favor edge pairs for which the closest
  points aren't endpoints of the edge.  So in addition to computing the
  distance between two edges, we compute the "subdistance," an integer
  0-2 which indicates how many of the closest points between the edge
  pair are endpoints.  When looking for the closest pair, we first 
  check the distance of a pair to the minimum distance found thus far.  If
  these distances are close, we check the subdistances.
*/

void closestEdges(f1, f2, T12, closest1, closest2)
     FACE *f1, *f2;
     MAT4 T12;
     EDGE **closest1, **closest2;
{
  FNODE *fn1, *fn2;
  REAL dist, minDist;
  int subDist, minSubDist;
  EDGE *e1, *e2;
  VECT3 cp1, cp2, tmp;

  minDist = INFINITY;
  for (fn1 = f1->edges; fn1; fn1 = fn1->next) {
    e1 = fn1->f.e;
    xformEdge(T12, e1);
    for (fn2 = f2->edges; fn2; fn2 = fn2->next) {
      e2 = fn2->f.e;
      edgeCPs(&e1, &e2, cp1, cp2);
      vectSub(cp2, cp1, tmp);
      dist = vectNorm(tmp);
      /* adjustment to favor edges for which closest pts aren't endpoints */
      subDist = (vectEqual(cp1, e1->v1->xcoords) || 
		 vectEqual(cp1, e1->v2->xcoords)) ? 1 : 0;
      if (vectEqual(cp2, e2->v1->coords) || vectEqual(cp2, e2->v2->coords))
	subDist++;
      if ((dist < minDist) || 
	  (dist < minDist + EPSILON && subDist < minSubDist)) {
	minDist = dist;
	minSubDist = subDist;
	*closest1 = e1;
	*closest2 = e2;
      }
    }
  }
}


/*
  Scan over the edges and vertices of face 2, and return the one which
  is closest to the plane of f1.  T12 & T21 are the usual transformation
  matrices.  Edges are favored over vertices when they are parallel to 
  the face.
*/

void *closestToFacePlane(f1, f2, T12, T21)
     FACE *f1, *f2;
     MAT4 T12, T21;
{
  FNODE *fn;
  EDGE *e;
  REAL dist, minDist, dist1, dist2;
  void *feat, *closest;
  VECT3 norm;

  minDist = INFINITY;
  xformVect(T12, f1->plane, norm);
  for (fn = f2->edges; fn; fn = fn->next) {
    e = fn->f.e;

    xformVert(T21, e->v1);
    dist1 = fabs(planeDist(f1->plane, e->v1->xcoords));
    xformVert(T21, e->v2);
    dist2 = fabs(planeDist(f1->plane, e->v2->xcoords));
    if (dist1 < dist2 - EPSILON) {
      /* v1 is closer */
      dist = dist1;
      feat = e->v1;
    }
    else if (dist2 < dist1 - EPSILON) {
      /* v2 is closer */
      dist = dist2;
      feat = e->v2;
    }
    else {
      /* edge is parallel to face */
      dist = (dist1 < dist2) ? dist1 : dist2;
      feat = e;
    }

    if ((dist < minDist) ||
	(dist < minDist + EPSILON && 
	 featTag(closest) == V && featTag(feat) == E)) {
      minDist = dist;
      closest = feat;
    }
  }
  return closest;
}
  

/*
  Determine if edge e intersects face f's cone.  Return true if it
  does, else false.  If true, min and max are parameter values which 
  bound the segment of e contained in the cone.  Note that
  0 <= min < max <= length(e).  We assume e has been xformed to
  f's frame.
*/

int polygonCut(f, e, min, max)
     FACE *f;
     EDGE *e;
     REAL *min, *max;
{
  PNODE *pn;
  REAL lambda;
  REAL denom;

  *min = 0.0;
  *max = e->len;

  for (pn = f->cone; pn; pn = pn->next) {
    denom = vectDotProd(e->xu, pn->plane);
    if (denom > 0.0) {
      lambda = - planeDist(pn->plane, e->v1->xcoords) / denom;
      if (lambda > *min) {
	*min = lambda;
	if (*min > *max) return 0;
      }
    }
    else if (denom < 0.0) {
      lambda = - planeDist(pn->plane, e->v1->xcoords) / denom;
      if (lambda < *max) {
	*max = lambda;
	if (*max < *min) return 0;
      }
    }
    else { /* denom = 0.0 ; lines are parallel */
      if (planeDist(pn->plane, e->v1->xcoords) < 0.0) return 0;
    }
  }

  return 1;
}


/*
  Return true if faces overlap, else false.  This is only called with
  pairs of parallel faces, and by "overlap" we mean the projection of
  one face onto the plane of the other overlaps the other face.

  We return true iff:
    some edge of f1 intersects the cone of f2, OR
    an abritrarily chosen vertex of f2 lies in the cone of f1
*/

int faceOverlap(f1, f2, T12, T21)
     FACE *f1, *f2;
     MAT4 T12, T21;
{
  FNODE *fn;
  REAL min, max;

  for (fn = f1->edges; fn; fn = fn->next) {
    xformEdge(T12, fn->f.e);
    if (polygonCut(f2, fn->f.e, &min, &max)) return 1;
  }

  /* no dice - maybe f2 lies completely within f1 */
  xformVert(T21, f2->verts->f.v);
  return faceConeChk(&f1, f2->verts->f.v->xcoords, 0);
}


/*
  =============================================================================

  feature pair checks and closest feature routine

  The six feature pair check routines test if a given pair of features
  is the closest pair of features between two polyhedra.  Each one handles
  a different combination of feature types.  The six routines are:
  vertex_vertex, vertex_edge, vertex_face, edge_edge, edge_face, and face_face.
  Each of these routines takes as parameters the two features as well as
  two transformation matrices for transforming features from one frame to
  the other and vice versa.  T12 is always the transformation matrix from
  the frame of the first feature in the parameter list to the frame of
  the second feature in the parameter list; T21 is the inverse transformation.
  The check routines which involve faces also take pointer(s) to the 
  polyhedron(a) of the face(s), since it is necessary to scan over the entire 
  polyhedron when a face's base plane is violated by a feature from the other 
  polyhedron.

  If pair of features passed into a check routine is indeed a closest
  feature pair, the distance between these features is returned.  If the
  features are not a closest feature pair, -INFINITY is returned, and
  one or possibly both of the closest features are updated to form a
  new pair which is guaranteed to be closer than the previous pair.

  The routine closestFeatures is the main loop of the distance algorithm.
  It iteratively "walks" around both polyhedra, calling the appropriate
  feature pair check routines, until it finds the closest pair of features.
  It then returns this pair.

  =============================================================================
*/


/*
  check if two vertices are a closest feature pair
*/

REAL vertex_vertex(v1, v2, T12, T21)
     VERTEX **v1, **v2;
     MAT4 T12, T21;
{
  VECT3 offset;

  /* check if v1 lies in v2's cone */
  xformVert(T12, *v1);
  if (!vertConeChk(v2, (*v1)->xcoords, 1)) return -INFINITY;

  /* check if v2 lies in v1's cone */
  xformVert(T21, *v2);
  if (!vertConeChk(v1, (*v2)->xcoords, 1)) return -INFINITY;

  vectSub((*v1)->coords, (*v2)->xcoords, offset);
  return vectNorm(offset);
}


/*
  check if a vertex and an edge are a closest feature pair
*/

REAL vertex_edge(v, e, T12, T21)
     VERTEX **v;
     EDGE **e;
     MAT4 T12, T21;
{
  VECT3 tmp, tmp2, offset;
  REAL lambda;

  /* check if v lies in e's cone */
  xformVert(T12, *v);
  if (!edgeConeChk(e, (*v)->xcoords, 1)) return -INFINITY;;

  /* compute closest point on e to v */
  vectSub((*v)->xcoords, (*e)->v1->coords, tmp);
  lambda = vectDotProd(tmp, (*e)->u);
  if (lambda < 0.0) lambda = 0.0;
  else if (lambda > (*e)->len) lambda = (*e)->len;
  displacePoint((*e)->v1->coords, (*e)->u, lambda, tmp);

  /* check if closest point on e lies in v's cone */
  xformPoint(T21, tmp, tmp2);
  if (!vertConeChk(v, tmp2, 1)) return -INFINITY;
  vectSub((*v)->coords, tmp2, offset);
  return vectNorm(offset);
}


/*
  check if two edges are a closest feature pair
*/

REAL edge_edge(e1, e2, T12, T21)
     EDGE **e1, **e2;
     MAT4 T12, T21;
{
  VECT3 cp1, cp2, offset, xformed;

  xformEdge(T12, *e1);
  edgeCPs(e1, e2, cp1, cp2);

  /* check if closest point on e1 lies in e2's cone */
  if (!edgeConeChk(e2, cp1, 1)) return -INFINITY;
  
  /* check if closest point on e2 lies in e1's cone */
  xformPoint(T21, cp2, xformed);
  if (!edgeConeChk(e1, xformed, 1)) return -INFINITY;

  vectSub(cp1, cp2, offset);
  return vectNorm(offset);
}


/*
  check if a vertex and a face are a closest feature pair
*/

REAL vertex_face(v, f, T12, T21, facePoly)
     VERTEX **v;
     FACE **f;
     MAT4 T12, T21;
     POLYHEDRON *facePoly;
{
  VECT3 tmp, tmp2;
  REAL dist;

  /* check if v lies in f's cone */
  xformVert(T12, *v);
  if (!faceConeChk(f, (*v)->xcoords, 1)) return -INFINITY;
  dist = planeDist((*f)->plane, (*v)->xcoords);
  if (dist < 0.0) {
    *f = closestToVert(*v, facePoly, T12);
    return -INFINITY;
  }

  /* 
    Compute closest point on f to v, i.e. proj. of v in f's plane.
    At this point, we know the projected point lies within f 
  */
  vectScale((*f)->plane, tmp, dist);
  vectSub((*v)->xcoords, tmp, tmp);


  /* check if closest point on f lies in v's cone */
  xformPoint(T21, tmp, tmp2);
  if (!vertConeChk(v, tmp2, 1)) return -INFINITY;

  return dist;
}

REAL edge_face(e, f, T12, T21, facePoly)
     EDGE **e;
     FACE **f;
     MAT4 T12, T21;
     POLYHEDRON *facePoly;
{
  PNODE *pn;
  VECT3 w, tmp;
  REAL h1, h2, absH1, absH2;
  REAL epsilon;
  VECT3 tmpV1, tmpV2;
  REAL min, max;

  /* compute distance of v1 and v2 from face */
  xformEdge(T12, *e);
  h1 = planeDist((*f)->plane, (*e)->v1->xcoords);
  h2 = planeDist((*f)->plane, (*e)->v2->xcoords);
  absH1 = fabs(h1);
  absH2 = fabs(h2);
  /* compute a variable epsilon which will prevent 
     edge-face / vertex-face cycling  and edge-face edge-edge cycling 
     (1.5 is a safety factor) */
  if (EDGE_F_FLARE > EDGE_V_FLARE /*VERT_FLARE*/)
    epsilon = /*1.5*/ 1.0e-4 + (*e)->len * sin(EDGE_F_FLARE);
  else epsilon = /*1.5*/ 1.0e-4 + (*e)->len * sin(EDGE_V_FLARE /*VERT_FLARE*/);

  /* fix for edges that would otherwise look || to f if h1 = -h2 */
  if (((h1 < 0.0 && h2 > 0.0) || (h1 > 0.0 && h2 < 0.0))
      && (fabs(h1 - h2) > epsilon)) {
    *f = (void *) closestEdgeOrVertOnFace(*f, *e);
    return -INFINITY;
  }
    
  

  if (absH1 < absH2 - epsilon) {    
    /* case I: v1 is closer */
    /* check if v1 lies in f's cone */
    for (pn = (*f)->cone; pn; pn = pn->next)
      if (planeDist(pn->plane, (*e)->v1->xcoords) < 0.0) break;
    if (pn) *f = (void *) closestEdgeOrVertOnFace(*f, *e);
    else if (h1 < 0.0) *f = closestToEdge(*e, facePoly, T12);
    else *e = (void *) (*e)->v1;
    return -INFINITY;
  }

  if (absH2 < absH1 - epsilon) {    
    /* case II: v2 is closer */
    /* check if v2 lies in f's cone */
    for (pn = (*f)->cone; pn; pn = pn->next)
      if (planeDist(pn->plane, (*e)->v2->xcoords) < 0.0) break;
    if (pn) *f = (void *) closestEdgeOrVertOnFace(*f, *e);
    else if (h2 < 0.0) *f = closestToEdge(*e, facePoly, T12);
    else *e = (void *) (*e)->v2;
    return -INFINITY;
  }

  else {   
    /* case III: e is parallel to face */

    if (!polygonCut(*f, *e, &min, &max)) {
      *f = (void *) closestEdgeOrVertOnFace(*f, *e);
      return -INFINITY;
    }

    displacePoint((*e)->v1->xcoords, (*e)->xu, min, tmpV1);
    h1 = planeDist((*f)->plane, tmpV1);
    displacePoint((*e)->v1->xcoords, (*e)->xu, max, tmpV2);
    h2 = planeDist((*f)->plane, tmpV2);
    if (h1 < 0.0 || h2 < 0.0) {
      *f = closestToEdge(*e, facePoly, T12);
      return -INFINITY;
    }

/*
  With flared planes, Ming's algo must be changed.  We must compare the
  face's normal directly to the two face plane normals of the edge's cone,
  rather than comparing it to a vector derived from the edge direction and
  the normals of the neighboring faces. 
  Note we assume the face planes are the first two in the edge's cone list!
*/
    xformVect(T21, (*f)->plane, w);
    if (vectDotProd(w, (*e)->cone->plane) > 0.0) {
      *e = (void *) (*e)->cone->nbr;
      return -INFINITY;
    }
    if (vectDotProd(w, (*e)->cone->next->plane) > 0.0) {
      *e = (void *) (*e)->cone->next->nbr;
      return -INFINITY;
    }

    if (h1 <= h2) {
      vectScale((*f)->plane, tmp, h1);
      return h1;
    }
    else {
      vectScale((*f)->plane, tmp, h2);
      return h2;
    }
  }
}


/*
  check if two faces are a closest feature pair

  The parallel, overlapping face-face case is rather complicated.  
  We first try to compute 
  the distance from face A to face B by finding the smallest positive dist of 
  any vertex of A from the plane of B or any vertex of B from the plane of A.
  Negative distances are ignored under the assumption that they are just due 
  to numerical errors that result when the vertex of A is very laterally 
  distant from B (e.g. inifinite planes), unless all distances are negative, 
  in which case we assume faces have interpenetrated.
*/


REAL face_face(f1, f2, T12, T21, face1poly, face2poly)
     FACE **f1, **f2;
     MAT4 T12, T21;
     POLYHEDRON *face1poly, *face2poly;
{
  PNODE *pn;
  FNODE *fn;
  EDGE *e1, *e2, *e;
  VECT3 w;
  REAL k;
  void *feat;
  REAL min, max;
  int penetration;
  REAL dist, minDist;
  VECT3 tmpV, tmpV2;

  xformVect(T21, (*f2)->plane, w);
  k = vectDotProd((*f1)->plane, w);
  if (fabs(k) > 1.0 - EPSILON) { 
    /* case I: faces are parallel */
    if (faceOverlap(*f1, *f2, T12, T21)) {

      /* Modified this case.  Used to determine if plane A was on (parallel)
	 plane B's inner side if the first vertex of A was on B's negative
	 side.  This cause probs when A is very large compared to B - even
	 if planes are roughly parallel, a distant vertex of A could appear
	 to lie quite far into B's negative side.  
	 
	 Fix: we clip one polygon to area of other one, therby
	 avoiding checks of vertices which are located far away from
	 one of the polygons. */

 
      minDist = INFINITY;
      penetration = 0;

      /* look at where f2's cone cuts f1's edges */
      for (fn = (*f1)->edges; fn; fn = fn->next) {
	e = fn->f.e;
	xformEdge(T12, e);
	if (polygonCut(*f2, e, &min, &max)) {
	  displacePoint(e->v1->xcoords, e->xu, min, tmpV);
	  dist = planeDist((*f2)->plane, tmpV);
	  if (dist < minDist) {
	    if (penetration = (dist < 0.0)) break;
	    minDist = dist;
	    vectScale((*f2)->plane, tmpV2, dist);
	  }
	  displacePoint(e->v1->xcoords, e->xu, max, tmpV);
	  dist = planeDist((*f2)->plane, tmpV);
	  if (dist < minDist) {
	    if (penetration = (dist < 0.0)) break;
	    minDist = dist;
	    vectScale((*f2)->plane, tmpV2, dist);
	  }
	}
      }
      if (penetration) {
	*f2 = closestToFace(*f1, face2poly, T12, T21);
	return -INFINITY;
      }

      /* look at where f1's cone cuts f2's edges */
      for (fn = (*f2)->edges; fn; fn = fn->next) {
	e = fn->f.e;
	xformEdge(T21, e);
	if (polygonCut(*f1, e, &min, &max)) {
	  displacePoint(e->v1->xcoords, e->xu, min, tmpV);
	  dist = planeDist((*f1)->plane, tmpV);
          if (dist < minDist) {
	    if (penetration = (dist < 0.0)) break;
	    minDist = dist;
	    vectScale((*f1)->plane, tmpV2, dist);
	  }
	  displacePoint(e->v1->xcoords, e->xu, max, tmpV);
	  dist = planeDist((*f1)->plane, tmpV);
	  if (dist <  minDist) {
	    if (penetration = (dist < 0.0)) break;
	    minDist = dist;
	    vectScale((*f1)->plane, tmpV2, dist);
	  }
	}
      }
      if (penetration) {
	*f1 = closestToFace(*f2, face1poly, T21, T12);
	return -INFINITY;
      }
      
      return minDist;
    }

    closestEdges(*f1, *f2, T12, &e1, &e2);
    *f1 = (void *) e1;
    *f2 = (void *) e2;
    return -INFINITY;
  }

  else {     
    /* case II: faces not parallel */ 

    /* return f2 & edge/vert of f1 */ 
    feat = closestToFacePlane(*f2, *f1, T21, T12);
    if (featTag(feat) == V) {
      xformVert(T12, (VERTEX *)feat);
      for (pn = (*f2)->cone; pn; pn = pn->next)
	if (planeDist(pn->plane, ((VERTEX *)feat)->xcoords) < 0.0) break;
      if (!pn) {
	*f1 = feat;
	return -INFINITY;
      }
    } 
    else {
      e = (EDGE *)feat;
      xformEdge(T12, e);
      /*
      h1 = planeDist((*f2)->plane, e->v1->xcoords);
      h2 = planeDist((*f2)->plane, e->v2->xcoords);
      vectScale((*f2)->plane, tmp, h1);
      vectSub(e->v1->xcoords, tmp, w1);
      vectScale((*f2)->plane, tmp, h2);
      vectSub(e->v2->xcoords, tmp, w2);
      if (polygonCut(*f2, w1, w2)) {
      */
      if (polygonCut(*f2, e, &min, &max)) {
	*f1 = feat;
	return -INFINITY;
      }
    }

    /* return f1 & edge/vert of f2 */ 
    feat = closestToFacePlane(*f1, *f2, T12, T21);
    if (featTag(feat) == V) {
      xformVert(T21, (VERTEX *)feat);
      for (pn = (*f1)->cone; pn; pn = pn->next)
	if (planeDist(pn->plane, ((VERTEX *)feat)->xcoords) < 0.0) break;
      if (!pn) {
	*f2 = feat;
	return -INFINITY;
      }
    } 
    else {
      e = (EDGE *)feat;
      xformEdge(T21, e);
      /*
      h1 = planeDist((*f1)->plane, e->v1->xcoords);
      h2 = planeDist((*f1)->plane, e->v2->xcoords);
      vectScale((*f1)->plane, tmp, h1);
      vectSub(e->v1->xcoords, tmp, w1);
      vectScale((*f1)->plane, tmp, h2);
      vectSub(e->v2->xcoords, tmp, w2);
      if (polygonCut(*f1, w1, w2)) {
      */
      if (polygonCut(*f1, e, &min, &max)) {
	*f2 = feat;
	return -INFINITY;
      }
    }
    closestEdges(*f1, *f2, T12, &e1, &e2);
    *f1 = (void *) e1;
    *f2 = (void *) e2;
    return -INFINITY;
  }
}


/* 
  try to get a distance between features, even if the closest
  feature algo is cycling.

  After the cycle detector is reset (by calling with reset = 1),
  cycleDetector() will repeatedly return -INFINITY, until a cycle
  is detected, at which time it will return a positive distance 
  - posisbly infinity - according to the smallest distance found so far.
  A cycle is detected when the distance between the current features
  fails to be lower than the minimum dist found thus far at least 5 times.
*/

REAL cycleDetector(feat1, feat2, T12, T21, reset)
     void **feat1, **feat2;
     MAT4 T12, T21;
     int reset; /* 1 to reset the cycle detection machinery */
{
  REAL dist;
  int flag;
  char name1[20], name2[20];

  static int cycleCounter;
  static REAL minDist;
  static void *bestCF1, *bestCF2;

  if (reset) {
    minDist = INFINITY;
    cycleCounter = 0;
    return 0.0;  /* return value irrelevant here */
  }

  featureName(*feat1, name1);
  featureName(*feat2, name2);
  printf("cycle? : feat1 = %-20s feat2 = %-s\n", name1, name2);

  flag = (featTag(*feat1) << 2) + featTag(*feat2);
  switch (flag) {
  case (V << 2) + V:
    xformVert(T21, *feat2);
    dist = Dvv(*feat1, *feat2);
    break;
  case (V << 2) + E:
    xformEdge(T21, *feat2);
    dist = Dve(*feat1, *feat2);
    break;
  case (V << 2) + F:
    xformVert(T12, *feat1);
    dist = Dfv(*feat2, *feat1);
    break;
  case (E << 2) + V:
    xformVert(T21, *feat2);
    dist = Dev(*feat1, *feat2);
    break;
  case (E << 2) + E:
    xformEdge(T21, *feat2);
    dist = Dee(*feat1, *feat2);
    break;
  case (E << 2) + F:
    xformEdge(T12, *feat1);
    dist = Dfe(*feat2, *feat1);
    break;
  case (F << 2) + V:
    xformVert(T21, *feat2);
    dist = Dfv(*feat1, *feat2);
    break;
  case (F << 2) + E:
    xformEdge(T21, *feat2);
    dist = Dfe(*feat1, *feat2);
    break;
  case (F << 2) + F:
    dist = Dff(*feat1, *feat2, T12, T21);
    break;
  }

  printf("dist=%+14.10f min=%+14.10f\n", dist, minDist);
  if (dist >= minDist) cycleCounter++;
  else {
    minDist = dist;
    bestCF1 = *feat1;
    bestCF2 = *feat2;
  }

  if (cycleCounter == 5) {
    /* cycle detected! */
    *feat1 = bestCF1;
    *feat2 = bestCF2;

    /* uncomment next line to find out how often cycles are occuring */
    /* printf("cycle detector returns %+12.6f\n", minDist); */

    return minDist;
  }

  else return -INFINITY;
}
    

/*
  closest feature pair

  Compute the closest feature pair for polyhedra poly1 & poly2.  The initial
  feature pair passed is passed in via feat1 & feat2, and upon termination
  these are updated to reflect the new closest feature pair.  The routine
  returns the distance between the pair of closest features.

  We keep track of the number of iterations of the algorithm (i.e. how
  many "steps" we take as we walk around the polyhedra, looking for
  closest features).  If this count reaches CYCLE_CNT, we assume we 
  are stuck in a cycle, and invoke the cycle detection system.
  Under normal circumstances, this should rarely occur.  If the polyhedra
  have high numbers of facets (tens of thousands) or experience large 
  changes in pose between calls to closestFeatures, CYCLE_CNT may have
  to be increased.
*/


REAL closestFeatures(poly1, feat1, poly2, feat2)
     POLYHEDRON *poly1, *poly2;
     void **feat1, **feat2;
{
  int type1, type2;
  REAL dist;
  MAT4 T12, T21, inv;
  int cycleChk;
#define CYCLE_CNT 500


  /* compute xformation matrices between the two polyhedron frames */
  matInvXform(poly2->pose, inv);
  matMultXform(inv, poly1->pose, T12);
  matInvXform(T12, T21);

  cycleChk = 0;

  do {

    
    if (cycleChk++ == CYCLE_CNT) cycleDetector(feat1, feat2, T12, T21, 1);
    else if (cycleChk > CYCLE_CNT) {
      dist = cycleDetector(feat1, feat2, T12, T21, 0);
      if (dist > 0.0) return dist;
    }


    type1 = featTag(*feat1);
    type2 = featTag(*feat2);

    if (type1 == V) {
      if (type2 == V) dist = vertex_vertex(feat1, feat2, T12, T21);
      else if (type2 == E) dist = vertex_edge(feat1, feat2, T12, T21);
      else dist = vertex_face(feat1, feat2, T12, T21, poly2);
    }
    else if (type1 == E) {
      if (type2 == V) dist = vertex_edge(feat2, feat1, T21, T12);
      else if (type2 == E) dist = edge_edge(feat1, feat2, T12, T21);
      else dist = edge_face(feat1, feat2, T12, T21, poly2);
    }
    else {
      if (type2 == V) dist = vertex_face(feat2, feat1, T21, T12, poly1);
      else if (type2 == E) 
	dist = edge_face(feat2, feat1, T21, T12, poly1);
      else dist = face_face(feat1, feat2, T12, T21, poly1, poly2);
    }

  } while (dist < 0.0);


  return dist;
}


/* 
  intialize closest feature pair

  Functionally, this is identical to closestFeatures, except we call this
  routine the first time, when we have not yet initialized feat1 and feat2.
  This routine simply chooses a starting feature on each of the polyhedra,
  and then calls closestFeatures.

  We choose the starting features in a fairly dumb manner, picking the
  first vertex in each polyhedron's vertex list.  This is totally arbitrary.
  There may be smarter ways of doing it, but it really doesn't matter, since
  we only perform this operation once, and then all subsequent calls are to
  closestFeatures.
*/

REAL closestFeaturesInit(poly1, feat1, poly2, feat2)
     POLYHEDRON *poly1, *poly2;
     void **feat1, **feat2;
{
  *feat1 = poly1->verts->f.v;
  *feat2 = poly2->verts->f.v;
  return closestFeatures(poly1, feat1, poly2, feat2);
}


    
