/*
  *****************************************************************************

  Copyright 1993 The Regents of the University of California.
  All Rights Reserved.

  Permission to use, copy, modify and distribute this software and its
  documentation for educational, research and non-profit purposes, without
  fee, and without a written agreement is hereby granted, provided that the 
  above copyright notice and the following three paragraphs appear in all 
  copies.

  Permission to incorporate this software into commercial products may be
  obtained from the Office of Technology Licensing, 2150 Shattuck Avenue,
  Suite 510, Berkeley, CA 94720-1620.

  IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
  DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
  LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION,
  EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY
  OF SUCH DAMAGES.

  THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
  AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
  ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATIONS
  TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.

  The author may be contacted via:

  US Mail:  Brian Mirtich
            387 Soda Hall
	    Computer Science Division
	    Department of Electrical Engineering and Computer Science
	    University of California
	    Berkeley, CA 94720

 Phone:     (510) 642-8149

 EMail:     mirtich@cs.berkeley.edu

 *****************************************************************************
*/



#include <stdio.h>
#include <math.h>
#include "matvect.h"



/*
  =============================================================================
  
  matrix stuff

  The following operations all deal with transformation matrices, 4 x 4 
  matrices corresponding to rigid body transformations of homogeneous
  coordinates.

  =============================================================================
*/


/* 4 x 4 identity matrix */

MAT4 mat4I = {{1, 0, 0, 0}, {0, 1, 0, 0}, {0, 0, 1, 0}, {0, 0, 0, 1}};


void mat4copy(source, dest)
     MAT4 source, dest;
{
  int i, j;

  for (i = 0; i < 4; i++)
    for (j = 0; j < 4; j++) dest[i][j] = source[i][j];
}


void mat4mult(a, b, c)
     MAT4 a, b, c;
{
  int i, j;

  for (i = 0; i < 4; i++)
    for (j = 0; j < 4; j++)
      c[i][j] = a[i][0] * b[0][j] + 
                a[i][1] * b[1][j] + 
                a[i][2] * b[2][j] + 
                a[i][3] * b[3][j];
}


/* 
  transformation matrix multiply:  a * b => c
  n.b. This routine is much faster than the general 4 x 4 matrix
  multiply above, but only works properly if a and b are SE(3) transformation
  matrices.  c should not point to the same matrix as a or b!
*/
void matMultXform(a, b, c)
     MAT4 a, b, c;
{
  int i, j;

  /* Rc = Ra Rb */
  c[0][0] = a[0][0]*b[0][0] + a[0][1]*b[1][0] + a[0][2]*b[2][0];
  c[0][1] = a[0][0]*b[0][1] + a[0][1]*b[1][1] + a[0][2]*b[2][1];
  c[0][2] = a[0][0]*b[0][2] + a[0][1]*b[1][2] + a[0][2]*b[2][2];
  c[1][0] = a[1][0]*b[0][0] + a[1][1]*b[1][0] + a[1][2]*b[2][0];
  c[1][1] = a[1][0]*b[0][1] + a[1][1]*b[1][1] + a[1][2]*b[2][1];
  c[1][2] = a[1][0]*b[0][2] + a[1][1]*b[1][2] + a[1][2]*b[2][2];
  c[2][0] = a[2][0]*b[0][0] + a[2][1]*b[1][0] + a[2][2]*b[2][0];
  c[2][1] = a[2][0]*b[0][1] + a[2][1]*b[1][1] + a[2][2]*b[2][1];
  c[2][2] = a[2][0]*b[0][2] + a[2][1]*b[1][2] + a[2][2]*b[2][2];

  /* Vc = Ra Vb + Va */
  c[0][3] = a[0][0]*b[0][3] + a[0][1]*b[1][3] + a[0][2]*b[2][3] + a[0][3];
  c[1][3] = a[1][0]*b[0][3] + a[1][1]*b[1][3] + a[1][2]*b[2][3] + a[1][3];
  c[2][3] = a[2][0]*b[0][3] + a[2][1]*b[1][3] + a[2][2]*b[2][3] + a[2][3];

  /* the rest */
  c[3][0] = c[3][1] = c[3][2] = 0.0;
  c[3][3] = 1.0;
}


/*
  transformation matrix inversion:  Inverse(M) => inv
  n.b. M and inv should not point to the same matrix.  We assume M is a
  transformation matrix; this will not properly invert arbitrary 4x4 matrices.
*/

void matInvXform(M, inv)
     MAT4 M, inv;
{
  /* we invert the rotation part by transposing it */
  inv[0][0] = M[0][0];
  inv[0][1] = M[1][0];
  inv[0][2] = M[2][0];
  inv[1][0] = M[0][1];
  inv[1][1] = M[1][1];
  inv[1][2] = M[2][1];
  inv[2][0] = M[0][2];
  inv[2][1] = M[1][2];
  inv[2][2] = M[2][2];

  /* the new displacement vector is given by:  d' = -(R^-1) * d */
  inv[0][3] = - inv[0][0]*M[0][3] - inv[0][1]*M[1][3] - inv[0][2]*M[2][3];
  inv[1][3] = - inv[1][0]*M[0][3] - inv[1][1]*M[1][3] - inv[1][2]*M[2][3];
  inv[2][3] = - inv[2][0]*M[0][3] - inv[2][1]*M[1][3] - inv[2][2]*M[2][3];

  /* the rest stays the same */
  inv[3][0] = inv[3][1] = inv[3][2] = 0.0;
  inv[3][3] = 1.0;
}

    
/*
  build xformation matrix
  
  axes is a string (e.g. "xyz", "yzy", or "zx") which specifies the order
  of axes rotations.  For example, "zx" means rotate about the z-axis, then
  about the x-axis.  All rotations are w.r.t. to the current frame, not a
  fixed frame.  angles[i] specifies the amount of rotation in degrees (!)
  about the axis specified by the ith character of axes.  dx, dy, and dz
  are the displacement components of the xformation matrix.  The matrix is
  returned in M.
*/

void matBuildXform(axes, angles, dx, dy, dz, M)
     char *axes;
     REAL angles[];
     REAL dx, dy, dz;
     MAT4 M;
{
  MAT4 Mi, Mold;
  REAL s, c;

  mat4copy(mat4I, M);
  for (; *axes; axes++) {
    mat4copy(M, Mold);
    mat4copy(mat4I, Mi);
    sincos(*angles * (M_PI / 180.0), &s, &c);

    angles++;
    if (*axes == 'x') {
      Mi[1][1] = Mi[2][2] = c;
      Mi[1][2] = -s;
      Mi[2][1] = s;
    }
    else if (*axes == 'y') {
      Mi[0][0] = Mi[2][2] = c;
      Mi[2][0] = -s;
      Mi[0][2] = s;
    }
    else {
      Mi[0][0] = Mi[1][1] = c;
      Mi[0][1] = -s;
      Mi[1][0] = s;
    }
    mat4mult(Mold, Mi, M);
  }
  M[0][3] = dx;
  M[1][3] = dy;
  M[2][3] = dz;
  M[3][0] = M[3][1] = M[3][2] = 0.0;
  M[3][3] = 1.0;
}



/*
  =============================================================================
  
  vector stuff

  The following operations manipulate vectors.  The vectors are assumed to
  be in R^3.

  =============================================================================
*/


/* 
  vector copy:  src => dest
*/

void vectCopy(src, dest)
     VECT3 src, dest;
{
  dest[0] = src[0];
  dest[1] = src[1];
  dest[2] = src[2];
}


/* 
  vector swap:  v1 <=> v2
*/

void vectSwap(v1, v2)
     VECT3 v1, v2;
{
  VECT3 tmp;

  vectCopy(v1, tmp);
  vectCopy(v2, v1);
  vectCopy(tmp, v2);
}


/* 
  vector addition:  a + b => c
*/

void vectAdd(a, b, c)
     VECT3 a, b, c;
{
  c[0] = a[0] + b[0];
  c[1] = a[1] + b[1];
  c[2] = a[2] + b[2];
}


/* 
  vector subtraction:  a - b => c
*/

void vectSub(a, b, c)
     VECT3 a, b, c;
{
  c[0] = a[0] - b[0];
  c[1] = a[1] - b[1];
  c[2] = a[2] - b[2];
}


/* 
  vector negation:  -src => dest
*/

void vectNeg(src, dest)
     VECT3 src, dest;
{
  dest[0] = - src[0];
  dest[1] = - src[1];
  dest[2] = - src[2];
}


/* 
  vector scaling: k * src => dest
*/

void vectScale(src, dest, k)
     VECT3 src, dest;
     REAL k;
{
  dest[0] = k * src[0];
  dest[1] = k * src[1];
  dest[2] = k * src[2];
}


/* 
  vector normalize: src / |src|  => dest
*/

void vectNormalize(src, dest)
     VECT3 src, dest;
{
  REAL l;

  l = vectNorm(src);
  dest[0] = src[0] / l;
  dest[1] = src[1] / l;
  dest[2] = src[2] / l;
}


/* 
  compute vector norm
*/

REAL vectNorm(v)
     VECT3 v;
{
  return sqrt(v[0]*v[0] + v[1]*v[1] + v[2]*v[2]);
}


/* 
  strict vector equality

  Return true if the vectors are exactly equal, else return false.
*/

int vectEqual(a, b)
     VECT3 a, b;
{
  return (a[0] == b[0] && a[1] == b[1] && a[2] == b[2]);
}


/*
  vector dot product:  return a . b
*/

REAL vectDotProd(a, b)
     VECT3 a, b;
{
  return a[0] * b[0] + a[1] * b[1] + a[2] * b[2];
}


/*
  vector cross product:  a x b => c
  n.b. c should not point to the same vector as a or b!
*/

void vectXprod(a, b, c)
     VECT3 a, b, c;
{
  c[0] = a[1] * b[2] - a[2] * b[1];
  c[1] = a[2] * b[0] - a[0] * b[2];
  c[2] = a[0] * b[1] - a[1] * b[0];
}


/*
  =============================================================================
  
  xformation stuff

  The following routines transform points and vectors using transformation
  matrices.  Points and vectors are both elements of R^3, but are 
  transformed differently.  Points are homogenized prior to transforming,
  meaning that a 1 is added as a fourth coordinate, before multiplying by
  the 4x4 transformation matrix.  For vectors, on the other hand, the
  fourth coordinate is 0.  The fourth coordinate of the transformed
  point or vector is never computed, since it is always 1 or 0
  respectively, and this value is implicitly assumed.

  =============================================================================
*/

    
/*
  transform a point:  M * (p 1) => (p2 1)
*/

void xformPoint(M, p, p2)
     MAT4 M;
     VECT3 p, p2;
{
  int i;

  for (i = 0; i < 3; i++) 
    p2[i] = M[i][0] * p[0] + M[i][1] * p[1] + M[i][2] * p[2] + M[i][3];
}


/*
  transform a vector:  M * (v 0) => (v2 0)
*/

void xformVect(M, v, v2)
     MAT4 M;
     VECT3 v, v2;
{
  int i;
  
  for (i = 0; i < 3; i++) 
    v2[i] = M[i][0] * v[0] + M[i][1] * v[1] + M[i][2] * v[2];
}
  
/*
  transform a 3-vector: M * x => x2
  This routine does a full transformation of a vector x in R^3.
*/

void xform3(M, x, x2)
     MAT3 M;
     VECT3 x, x2;
{
  int i;

  for (i = 0; i < 3; i++) 
    x2[i] = M[i][0] * x[0] + M[i][1] * x[1] + M[i][2] * x[2];
}



/*
  =============================================================================
  
  misc. stuff

  The following are miscellaneous operations for points, vectors, etc.

  =============================================================================
*/



/* 
  plane distance

  Compute the distance from a point to a plane.  The plane is specified
  by four coefficients, i.e. the plane equation is

       plane[0] * x + plane[1] * y + plane[2] * z + plane[3] = 0.

  We assume that the first three plane coordinates form a unit vector
  (the unit normal to the plane).  The "distance" returned is actually
  a signed distance.  That is, if the point lies on the side of the plane
  to which the unit normal points, the result is positive.  If the point
  lies on the other side, the distance is negative.
*/

REAL planeDist(plane, point)
     VECT4 plane;
     VECT3 point;
{
  return plane[0] * point[0] + 
         plane[1] * point[1] + 
         plane[2] * point[2] + plane[3];
}


/* 
  displace point:  point + lambda * vect => result

  Compute the point which is displaced (lambda * vect) from point.
*/


void displacePoint(point, vect, lambda, result)
     VECT3 point;
     VECT3 vect;
     REAL lambda;
     VECT3 result;
{
  result[0] = point[0] + lambda * vect[0];
  result[1] = point[1] + lambda * vect[1];
  result[2] = point[2] + lambda * vect[2];
}

